# -*- coding: utf-8 -*-
from datetime import datetime
from decimal import Decimal
from typing import Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, computed_field

from vpnflow.settings import settings


class BaseSchemaModel(BaseModel):
    """⭐"""

    model_config = ConfigDict(from_attributes=True, validate_assignment=True, populate_by_name=True)


class User(BaseSchemaModel):
    """⭐"""
    id: Optional[int] = None
    ts_created: Optional[datetime] = None
    ts_updated: Optional[datetime] = None
    telegram_id: int
    marzban_username: Optional[str] = None
    payplan_id: Optional[int] = None
    referral_id: Optional[int] = None


class UserTelegram(BaseSchemaModel):
    """⭐"""
    id: Optional[int] = None
    name: Optional[str] = None
    name_full: Optional[str] = None
    language_code: Optional[str] = None
    is_bot: Optional[bool] = None
    is_premium: Optional[bool] = None
    referral_code: Optional[str] = None


class Referral(BaseSchemaModel):
    """⭐"""
    id: Optional[int] = None
    telegram_start: str
    telegram_id: Optional[int] = None


class PayPlan(BaseSchemaModel):
    """⭐"""
    id: Optional[int] = None
    name: str
    days_duration: int
    data_limit_gb: int
    data_limit_reset_strategy: str
    is_active: Optional[bool] = None


class PayPlanPrice(BaseSchemaModel):
    """⭐"""
    payplan_id: int
    payment_method: str
    price: Decimal
    currency_code: str


class PayPlanPriceCurrency(BaseSchemaModel):
    """⭐"""
    code: str
    symbol: str


class PayPlanPriceView(PayPlanPrice):
    """⭐"""
    currency: PayPlanPriceCurrency

    @computed_field
    def price_short(self) -> int:
        """⭐"""
        return int(self.price)


class PayPlanView(PayPlan):
    """⭐"""
    prices: Optional[List[PayPlanPriceView]] = []

    @computed_field
    def button_text(self) -> str:
        """⭐"""
        prices = ' | '.join(
            (f"{int(price.price)} {price.currency.symbol or price.currency_code}" for price in self.prices)
            )
        months = self.days_duration // 30
        if months > 1:
            prices_per_month = '/'.join(
                (f"{int(price.price / months)}{price.currency.symbol or price.currency_code}" for price in self.prices)
                )
            return f"{self.name} - {prices}\n({prices_per_month}/мес.)"
        return f"{self.name} - {prices}"

    @computed_field
    def data_limit(self) -> int:
        """⭐"""
        return 2 ** 30 * self.data_limit_gb


class Coupon(BaseSchemaModel):
    """⭐"""
    id: str
    days_duration: int
    is_active: Optional[bool] = None
    usage_limit: Optional[int] = None


class UsageCoupon(BaseSchemaModel):
    """⭐"""
    user_id: int
    coupon_id: str


class Payment(BaseSchemaModel):
    """⭐"""
    id: Optional[UUID] = None
    ts_created: Optional[datetime] = None
    ts_updated: Optional[datetime] = None
    payplan_id: int
    payment_method: str
    price: int
    currency: str
    paid: Optional[bool] = None
    title: str
    telegram_payment_charge_id: Optional[str] = None
    provider_payment_charge_id: Optional[str] = None


class UserPayment(BaseSchemaModel):
    """⭐"""
    user_id: int
    payment_id: UUID


class UserBalance(BaseSchemaModel):
    """⭐"""
    id: Optional[int] = None
    user_id: int
    currency: str
    amount: Optional[Decimal] = None


class UserBalanceView(UserBalance):
    """⭐"""
    symbol: Optional[str] = None

    @computed_field
    def text(self) -> str:
        """⭐"""
        return f"{int(self.amount)} {self.symbol or self.currency}"


class ReferralPayment(BaseSchemaModel):
    """⭐"""
    ts_created: Optional[datetime] = None
    user_id: int
    referral_id: int
    amount: Decimal
    currency: str


class MarzbanUserCreate(BaseSchemaModel):
    """⭐"""
    username: str
    note: Optional[str] = ""
    proxies: Dict
    data_limit: int
    expire: int
    data_limit_reset_strategy: str = "no_reset"
    inbounds: Dict


class MarzbanResponse(BaseSchemaModel):
    """⭐"""

    model_config = ConfigDict(extra="allow")


class MarzbanUserCreated(MarzbanResponse):
    """⭐"""
    username: str
    links: List[str]
    subscription_url: str


class YookassaProviderDataReceiptItem(BaseSchemaModel):
    """⭐"""
    description: str
    quantity: int = 1
    amount: Dict
    vat_code: int = 1
    payment_mode: str = "full_payment"
    payment_subject: str = "commodity"


class YookassaProviderDataReceipt(BaseSchemaModel):
    """⭐"""
    items: List[YookassaProviderDataReceiptItem]
    tax_system_code: int = 1


class YookassaProviderData(BaseSchemaModel):
    """⭐"""
    receipt: YookassaProviderDataReceipt


class CloudFlareDnsRecordCreate(BaseSchemaModel):
    """⭐"""
    name: str
    content: str
    type: Optional[str] = "A"
    comment: Optional[str] = "Added from telegram bot"
    proxied: Optional[bool] = False


class MarzbanNodeCreate(BaseSchemaModel):
    """⭐"""
    name: str
    address: str
    port: Optional[int] = settings.marzban.port
    api_port: Optional[int] = settings.marzban.api_port
    add_as_new_host: Optional[bool] = True
    usage_coefficient: Optional[int] = 1



# -*- coding: utf-8 -*-
import os
from contextlib import asynccontextmanager
from logging import getLogger

from aiogram.exceptions import TelegramRetryAfter
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from prometheus_fastapi_instrumentator import Instrumentator

from vpnflow.bot.base import create_dispatcher, on_shutdown_bot, on_startup_bot
from vpnflow.settings import settings
from vpnflow.web.views import ROUTERS

logger = getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI):
    """⭐"""
    pid = os.getpid()
    app.settings = settings
    app.bot_dispatcher = create_dispatcher()
    app.bot = app.bot_dispatcher.bot
    try:
        logger.info(f"Webserver startup: {pid}")
        await on_startup_bot(app.bot, app.bot_dispatcher)
        yield
        logger.info(f"Webserver on_shutdown: {pid}")
        await on_shutdown_bot(app.bot, app.bot_dispatcher)
    except TelegramRetryAfter as exc:
        logger.error(exc)
    except Exception as exc:
        logger.error(f"{type(exc)}. {exc}")
    finally:
        pass


app = FastAPI(**{**settings.webserver.dict(), **{"lifespan": lifespan}})
if settings.webserver.static_url and os.path.exists(settings.webserver.static_dir):
    app.mount(
        settings.webserver.static_url,
        StaticFiles(directory=settings.webserver.static_dir), name="static"
        )
if settings.webserver.templates_dir and os.path.exists(settings.webserver.templates_dir):
    app.templates = Jinja2Templates(directory=settings.webserver.templates_dir)
for router in ROUTERS:
    app.include_router(router)
Instrumentator().instrument(app).expose(app)


@app.middleware("http")
async def exception_handling(request: Request, call_next):
    """⭐"""
    try:
        return await call_next(request)
    except Exception as exc:
        logger.error(exc)
        return JSONResponse({"message": "error"}, status_code=500)



# -*- coding: utf-8 -*-
from asyncio import create_task
from datetime import datetime
from logging import getLogger
from typing import Final

from aiogram import Bot, F, Router
from aiogram.enums import ChatType
from aiogram.filters import Command, CommandObject, CommandStart
from aiogram.fsm.context import FSMContext
from aiogram.types import CallbackQuery, Message
from aiogram.types.pre_checkout_query import PreCheckoutQuery

from vpnflow.bot import callbacks as cb
from vpnflow.bot import keyboards as kbs
from vpnflow.bot import states as st
from vpnflow.bot.filters import BlacklistFilter
from vpnflow.bot.middleware import CheckAcceptMiddleware
from vpnflow.db.models import User
from vpnflow.enums import BotCommands
from vpnflow.misc import days_verbose
from vpnflow.services import _telegram, business
from vpnflow.settings import settings
from vpnflow.web import schemas

router: Final[Router] = Router(name=__name__)
router.message.middleware(CheckAcceptMiddleware())
router.callback_query.middleware(CheckAcceptMiddleware())
router.message.filter(BlacklistFilter(), F.chat.type == ChatType.PRIVATE)
router.callback_query.filter(BlacklistFilter(), F.message.chat.type == ChatType.PRIVATE)
logger = getLogger(__name__)
messages = settings.telegram.messages
business_settings = settings.business


@router.message(CommandStart(deep_link=True))
@router.message(Command(BotCommands.start.name))
async def cmd_start(message: Message, command: CommandObject) -> None:
    """⭐"""
    user_data = message.from_user
    is_exists = await business.is_telegram_user_exists(user_data.id)
    if not is_exists:
        user_created = await business.create_user_from_telegram(user_data, command.args)
        if user_created:
            await message.answer(messages["start-welcome"], reply_markup=kbs.inline.user_welcome)
        return
    user = await business.get_user(telegram_id=user_data.id)
    if user is None:
        logger.warning(f"User is None, {user_data.id}")
        return
    if user and user.marzban_username is None:
        await message.answer(messages["start-welcome"], reply_markup=kbs.inline.user_welcome)
        return
    user_balances = await business.get_user_balances(user)
    marzban_user = await business.get_user_marzban(user.marzban_username)
    msg_id = "start-success" if marzban_user.status == "active" else "start-expired"
    await message.answer(
        messages[msg_id].format(
            marzban_username=user.marzban_username, user_balance=user_balances,
            sub_date=datetime.fromtimestamp(marzban_user.expire).strftime("%d.%m.%Y"),
            subscription_url=marzban_user.subscription_url
            ),
        reply_markup=kbs.inline.user,
        disable_web_page_preview=True
        )
    return


@router.callback_query(cb.UserCallback.filter(F.action == "sign_up"))
async def cb_sign_up(
    callback_query: CallbackQuery, callback_data: cb.UserCallback
    ) -> None:
    """⭐"""
    user = await business.get_user(telegram_id=callback_query.from_user.id)
    if not user:
        return
    test_days = business_settings.test_days
    test_traffic_gb_limit = business_settings.test_traffic_gb_limit
    test_proxies = business_settings.test_proxies
    test_inbounds = business_settings.test_inbounds
    marzban_user = await business.create_user_marzban(
        days=test_days, traffic_limit=test_traffic_gb_limit,
        proxies=test_proxies, inbounds=test_inbounds
        )
    _ = await business.update_user(user, marzban_username=marzban_user.username)
    user = await business.get_user(telegram_id=callback_query.from_user.id)
    await _telegram.edit_callback(
        callback_query,
        messages["start-new"].format(days=test_days),
        reply_markup=kbs.inline.user_start_new
        )


@router.callback_query(cb.UserCallback.filter(F.action == "start"))
async def cb_start(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    await _telegram.clean_state(state)
    user_balances = await business.get_user_balances(user)
    marzban_user = await business.get_user_marzban(user.marzban_username)
    msg_id = "start-success" if marzban_user.status == "active" else "start-expired"
    await _telegram.edit_callback(
        callback_query,
        messages[msg_id].format(
            marzban_username=user.marzban_username, user_balance=user_balances,
            sub_date=datetime.fromtimestamp(marzban_user.expire).strftime("%d.%m.%Y"),
            subscription_url=marzban_user.subscription_url
            ),
        reply_markup=kbs.inline.user
        )


@router.message(Command(BotCommands.pay.name))
async def cmd_pay(message: Message, state: FSMContext, user: User) -> None:
    """⭐"""
    await message.answer(
        messages[BotCommands.pay.name],
        reply_markup=await kbs.inline.build_pay_plans(business.get_pay_plans())
        )
    await state.set_state(st.PayPlanForm.payplan_choice)


@router.callback_query(cb.UserCallback.filter(F.action == "pay"))
async def cb_pay(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    await _telegram.edit_callback(
        callback_query, messages[BotCommands.pay.name],
        reply_markup=await kbs.inline.build_pay_plans(business.get_pay_plans())
        )
    await state.set_state(st.PayPlanForm.payplan_choice)

@router.callback_query(
    cb.PayPlanCallback.filter(F.action == "choice-pay-plan"),
    st.PayPlanForm.payplan_choice
    )
async def cb_choice_pay_plan(
    callback_query: CallbackQuery,
    callback_data: cb.PayPlanCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    payplan_id = callback_data.value
    await state.update_data(payplan_id=payplan_id)
    pay_plan = await business.get_pay_plan(payplan_id)

    await state.update_data(view_days_duration=pay_plan.days_duration)

    if len(pay_plan.prices) == 1:
        pay_plan_price = pay_plan.prices[0]
        await state.update_data(payment_method=pay_plan_price.payment_method)

        user_has_balance = await business.check_user_balance(
            callback_query.from_user.id, pay_plan_price.currency_code, pay_plan_price.price
            )
        if user_has_balance:
            user = await business.get_user(telegram_id=callback_query.from_user.id)
            user_balances = await business.get_user_balances(user)
            await _telegram.edit_callback(
                callback_query,
                messages["pay-plan-balance"].format(
                    pay_plan_name=pay_plan.name,
                    pay_plan_price=pay_plan_price.price_short,
                    pay_plan_currency_symbol=pay_plan_price.currency.symbol,
                    user_balance=user_balances
                    ), reply_markup=kbs.inline.yes_no_menu
                )
            await state.update_data(
                pay_plan_name=pay_plan.name,
                pay_plan_price=pay_plan_price.price_short,
                pay_plan_currency_symbol=pay_plan_price.currency.symbol
                )
            await state.set_state(st.PayPlanForm.write_off_balance)
        else:
            await _telegram.edit_callback(
                callback_query,
                messages["pay-plan"].format(
                    pay_plan_name=pay_plan.name,
                    pay_plan_price=pay_plan_price.price_short,
                    pay_plan_currency_symbol=pay_plan_price.currency.symbol
                    ), reply_markup=kbs.inline.yes_no_menu
                )
            await state.set_state(st.PayPlanForm.accept_choice)
        return

    await _telegram.edit_callback(
        callback_query, messages["pay-method"],
        reply_markup=kbs.inline.build_pay_plan_prices(pay_plan)
        )
    await state.set_state(st.PayPlanForm.pay_method_choice)


@router.callback_query(
    cb.PayPlanCallback.filter(F.action == "choice-pay-method"),
    st.PayPlanForm.pay_method_choice
    )
async def cb_choice_pay_method(
    callback_query: CallbackQuery,
    callback_data: cb.PayPlanCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    pay_method = callback_data.value
    await state.update_data(payment_method=pay_method)
    data = await state.get_data()
    pay_plan = await business.get_pay_plan(data.get("payplan_id"))
    for pay_plan_price in pay_plan.prices:
        if pay_plan_price.payment_method == pay_method:
            user_has_balance = await business.check_user_balance(
                callback_query.from_user.id, pay_plan_price.currency_code,
                pay_plan_price.price
                )
            if user_has_balance:
                await _telegram.edit_callback(
                    callback_query,
                    messages["pay-plan-balance"].format(
                        pay_plan_name=pay_plan.name,
                        pay_plan_price=pay_plan_price.price_short,
                        pay_plan_currency_symbol=pay_plan_price.currency.symbol
                    )
                    , reply_markup=kbs.inline.yes_no_menu)
                await state.update_data(
                    pay_plan_name=pay_plan.name,
                    pay_plan_price=pay_plan_price.price_short,
                    pay_plan_currency_symbol=pay_plan_price.currency.symbol
                    )
                await state.set_state(st.PayPlanForm.write_off_balance)
            else:
                await _telegram.edit_callback(
                    callback_query,
                    messages["pay-plan"].format(
                        pay_plan_name=pay_plan.name,
                        pay_plan_price=pay_plan_price.price_short,
                        pay_plan_currency_symbol=pay_plan_price.currency.symbol
                    )
                    , reply_markup=kbs.inline.yes_no_menu)
                await state.set_state(st.PayPlanForm.accept_choice)
            return


@router.callback_query(
    cb.YesNoCallback.filter(),
    st.PayPlanForm.write_off_balance
    )
async def cb_write_off_balance(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    state_data = await state.get_data()
    if callback_data.value:
        services_access = await business.check_services_access()
        if not services_access:
            logger.error("Some services are not available")
            return
        payment = await business.create_payment(callback_query.from_user.id, **state_data)
        _ = await business.update_user_after_pay_no_provider(callback_query.from_user.id, payment.id)
        await callback_query.message.answer(
            messages["pay-success"].format(days_duration=state_data["view_days_duration"]),
            reply_markup=kbs.inline.user_start
            )
        await _telegram.clean_state(state)
    else:
        await _telegram.edit_callback(
            callback_query,
            messages["pay-plan"].format(**state_data),
            reply_markup=kbs.inline.yes_no_menu
            )
        await state.set_state(st.PayPlanForm.accept_choice)


@router.callback_query(
    cb.YesNoCallback.filter(),
    st.PayPlanForm.accept_choice
    )
async def cb_accept_choice(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    state_data = await state.get_data()
    await _telegram.clean_state(state)
    if callback_data.value:
        payment = await business.create_payment(callback_query.from_user.id, **state_data)
        await _telegram.send_invoice(callback_query, payment)
        await state.update_data(view_days_duration=state_data["view_days_duration"])
    else:
        await callback_query.answer(text=messages["pay-no"]) # show_alert=True
        await _telegram.edit_callback(
            callback_query, messages[BotCommands.pay.name],
            reply_markup=await kbs.inline.build_pay_plans(business.get_pay_plans())
            )
        await state.set_state(st.PayPlanForm.payplan_choice)


@router.pre_checkout_query()
async def payment_pre_checkout_query(pre_checkout_query: PreCheckoutQuery) -> None:
    """⭐"""
    try:
        services_access = await business.check_services_access()
        if not services_access:
            raise Exception("Some services are not available")
        await pre_checkout_query.answer(ok=True)
    except Exception as exc:
        logger.error(exc)
        await pre_checkout_query.answer(
            ok=False,
            error_message=messages["pay-failed"].format(pay_id=pre_checkout_query.id)
            )


@router.message(F.successful_payment)
async def payment_successful(message: Message, state: FSMContext, bot: Bot) -> None:
    """⭐"""
    state_data = await state.get_data()
    await _telegram.clean_state(state)
    user_id, payment_info = message.from_user.id, message.successful_payment
    user_payment = await business.update_user_after_pay(
        user_id=user_id, payment_info=payment_info, bot=bot
        )
    if not user_payment:
        logger.error(f"Invoice paid, but profile not updated user id: {user_id}, payment_info: {payment_info}")
        await message.answer(messages["error"], reply_markup=kbs.inline.user_start)
        return
    await message.answer(
        messages["pay-success"].format(days_duration=state_data["view_days_duration"]),
        reply_markup=kbs.inline.user_start
        )
    create_task(
        _telegram.broadcast(
            bot,
            messages["admin-alert-pay"].format(
                marzban_username=user_payment.user.marzban_username,
                price=user_payment.payment.price,
                currency=user_payment.payment.currency
                ),
            *settings.telegram.admins_id
            )
        )


@router.callback_query(cb.UserCallback.filter(F.action == "setup"))
async def cb_setup(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    await _telegram.edit_callback(
        callback_query, messages["setup-platform"], reply_markup=kbs.inline.user_setup_platform
        )


@router.callback_query(cb.UserCallback.filter(F.action == "setup-platform"))
async def cb_setup_platform(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    marzban_user = await business.get_user_marzban(user.marzban_username)
    if marzban_user is None:
        return
    download_url = business_settings.supported_platforms[callback_data.value]
    add_url = business.define_app_by_url(download_url)
    if add_url is None:
        return
    add_url = business_settings.url_redirect + add_url.format(subscription_url=marzban_user.subscription_url)
    await _telegram.edit_callback(
        callback_query, messages["setup-steps"],
        reply_markup=kbs.inline.build_setup_steps(download_url, add_url)
        )


@router.callback_query(cb.UserCallback.filter(F.action == "setup-help"))
async def cb_setup_help(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    marzban_user = await business.get_user_marzban(user.marzban_username)
    if marzban_user is None:
        return
    await _telegram.edit_callback(
        callback_query, messages["setup-help"].format(subscription_url=marzban_user.subscription_url),
        reply_markup=kbs.inline.support_back
        )


@router.callback_query(cb.UserCallback.filter(F.action == "promo"))
async def cb_promo(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    state: FSMContext, user: User
    ) -> None:
    """⭐"""
    _ = await business.get_user_marzban(user.marzban_username)
    await _telegram.edit_callback(callback_query, messages["promo-success"])
    await state.set_state(st.CouponForm.coupon)


@router.message(st.CouponForm.coupon)
async def promo_coupon(message: Message, state: FSMContext) -> None:
    """⭐"""
    await _telegram.clean_state(state)
    user = await business.get_user(telegram_id=message.from_user.id)
    filters = schemas.UsageCoupon(
        user_id=user.id, coupon_id=message.text.strip().upper()
        ).dict()
    coupon, coupon_usage, coupon_breaked_limit = await business.get_coupon_and_coupon_usage(**filters)
    if coupon is None:
        await message.answer(messages["promo-coupon-not-exists"], reply_markup=kbs.inline.user_promo)
        return
    if coupon_usage is not None:
        await message.answer(messages["promo-coupon-is-used"], reply_markup=kbs.inline.user_promo)
        return
    if coupon_breaked_limit:
        await message.answer(messages["promo-coupon-is-limit"], reply_markup=kbs.inline.user_promo)
        return
    coupon_usage_created = await business.create_coupon_usage(**filters)
    updated_user_marzban = await business.update_user_marzban(
        user.marzban_username, coupon.days_duration
        )
    if coupon_usage_created and updated_user_marzban:
        await message.answer(
            messages["promo-coupon"].format(
                days_duration=coupon.days_duration, days_say=days_verbose(coupon.days_duration)
                )
            )
        user_balances = await business.get_user_balances(user)
        marzban_user = await business.get_user_marzban(user.marzban_username)
        msg_id = "start-success" if marzban_user.status == "active" else "start-expired"
        await message.answer(
            messages[msg_id].format(
                marzban_username=user.marzban_username, user_balance=user_balances,
                sub_date=datetime.fromtimestamp(marzban_user.expire).strftime("%d.%m.%Y"),
                subscription_url=marzban_user.subscription_url
                ),
            reply_markup=kbs.inline.user,
            disable_web_page_preview=True
            )
        return


@router.callback_query(cb.UserCallback.filter(F.action == "invite"))
async def cb_invite(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    user: User
    ) -> None:
    """⭐"""
    user_balances = await business.get_user_balances(user)
    invite_url = _telegram.create_ref_link(await business.get_referral_code_by_telegram_id(callback_query.from_user.id))
    invited_users = await business.get_invited_users(callback_query.from_user.id)
    referral_balance = await business.get_referral_balance(user.id)
    await _telegram.edit_callback(
        callback_query,
        messages["invite"].format(
            user_balance=user_balances, referral_balance=referral_balance,
            invite_url=invite_url, invited_users=invited_users,
            referral_percent=business_settings.referral_percent_text
            ),
        reply_markup=kbs.inline.build_referral(invite_url)
        )


@router.message(Command(BotCommands.help.name))
async def cmd_help(message: Message, user: User) -> None:
    """⭐"""
    await message.answer(messages["help"], reply_markup=kbs.inline.support)


@router.callback_query(cb.UserCallback.filter(F.action == "help"))
async def cb_help(
    callback_query: CallbackQuery, callback_data: cb.UserCallback,
    user: User
    ) -> None:
    """⭐"""
    await _telegram.edit_callback(callback_query, messages["help"], reply_markup=kbs.inline.support)


@router.message(Command(BotCommands.ref.name))
async def cmd_ref(message: Message, state: FSMContext, user: User) -> None:
    """⭐"""
    user_balances = await business.get_user_balances(user)
    invite_url = _telegram.create_ref_link(await business.get_referral_code_by_telegram_id(message.from_user.id))
    invited_users = await business.get_invited_users(message.from_user.id)
    referral_balance = await business.get_referral_balance(user.id)
    await message.answer(
        messages["invite"].format(
            user_balance=user_balances, referral_balance=referral_balance,
            invite_url=invite_url, invited_users=invited_users,
            referral_percent=business_settings.referral_percent_text
            ),
        reply_markup=kbs.inline.build_referral(invite_url)
        )



# -*- coding: utf-8 -*-
import secrets
import string
from logging import getLogger
from logging.config import dictConfig

import yaml

logger = getLogger(__name__)


def load_yaml(file_path: str) -> dict:
    """⭐"""
    with open(file_path, 'r', encoding="UTF-8") as f:
        return yaml.safe_load(f.read())


def load_log_conf(file_path: str) -> None:
    """⭐"""
    dictConfig(load_yaml(file_path))


def days_verbose(days: int) -> str:
    """⭐"""
    if days > 20:
        days %= 10
    if days == 1:
        return "день"
    elif 2 <= days <= 4:
        return "дня"
    return "дней"


def slice_generator(s, l):
    """⭐"""
    for i in range(0, len(s), l):
        yield s[i:i + l]


def generate_symbol_sequence(size: int = 5, seq: str = string.ascii_letters + string.digits) -> str:
    """⭐"""
    return "".join(secrets.choice(seq) for _ in range(size))



# -*- coding: utf-8 -*-
from typing import Any, Optional, Union

from aiogram.filters.callback_data import CallbackData


class PayPlanCallback(CallbackData, prefix="payplan"):
    """⭐"""
    action: str
    value: Union[int, str]


class YesNoCallback(CallbackData, prefix="yesno"):
    """⭐"""
    value: bool


class AdminCallback(CallbackData, prefix="admin"):
    """⭐"""
    action: str


class SupportCallback(CallbackData, prefix="support"):
    """⭐"""
    action: str


class UserCallback(CallbackData, prefix="user"):
    """⭐"""
    action: str
    value: Optional[Any] = None


class NotifyCallback(CallbackData, prefix="notify"):
    """⭐"""
    value: str



# -*- coding: utf-8 -*-
__author__ = "Michael R. Kisel"
__copyright__ = "Michael R. Kisel"
__license__ = "AGPL"
__version__ = "1.0.1"
__maintainer__ = "Michael R. Kisel"
__email__ = "aioboy@yandex.com"
__status__ = "Stable"



# -*- coding: utf-8 -*-
from dagster import (DefaultScheduleStatus, Definitions, RetryPolicy,
                     ScheduleDefinition, job, op)

from vpnflow.flows.base import notify_users
from vpnflow.settings import settings


@op(
    retry_policy=RetryPolicy(
        max_retries=settings.flows.notify_users_retry_max,
        delay=settings.flows.notify_users_retry_delay
        )
    )
def op_notify_users():
    """⭐"""
    notify_users()


@job
def job_notify_users():
    """⭐"""
    op_notify_users()


schedule_notify_users = ScheduleDefinition(
    cron_schedule=settings.flows.notify_users_cron, name="notify_users",
    description="Notify in telegram soon expired users",
    job=job_notify_users, execution_timezone="UTC",
    default_status=DefaultScheduleStatus.RUNNING
)


definitions = Definitions(
    schedules=[schedule_notify_users],
    jobs=[job_notify_users]
)



# -*- coding: utf-8 -*-
import logging

__all__ = ("MarkdownFormatter", "HtmlFormatter")


class BaseFormatter(logging.Formatter):
    """⭐"""

    EMOJI_MAP = {
        logging.CRITICAL: '⚫',
        logging.ERROR: '🔴',
        logging.WARNING: '🟠',
        logging.INFO: '🔵',
        logging.DEBUG: '⚪',
        logging.NOTSET: '⚪'
    }

    def __init__(self, *args, **kwargs):
        """⭐"""
        super().__init__(*args, **kwargs)

    @staticmethod
    def escape_html(text):
        """⭐"""
        return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

    def format(self, record):
        """⭐"""
        super().format(record)
        if record.funcName:
            record.funcName = self.escape(str(record.funcName))
        if record.name:
            record.name = self.escape(str(record.name))
        if record.msg:
            record.msg = self.escape(record.getMessage())
        emoji = self.EMOJI_MAP[record.levelno]
        record.levelname = emoji + " #" + record.levelname
        return self._style.format(record)

    def formatException(self, *args, **kwargs):
        """⭐"""
        return f"{self.BLOCK_L}{self.escape(self.formatException(*args, **kwargs))}{self.BLOCK_R}"

    def formatStack(self, *args, **kwargs):
        """⭐"""
        return f"{self.BLOCK_L}{self.escape(self.formatStack(*args, **kwargs))}{self.BLOCK_R}"


class MarkdownFormatter(BaseFormatter):
    """⭐"""

    BLOCK_L = BLOCK_R = "```"

    @staticmethod
    def escape(s):
        """⭐"""
        return s


class HtmlFormatter(BaseFormatter):
    """⭐"""

    BLOCK_L, BLOCK_R = "<pre>", "</pre>"

    @staticmethod
    def escape(s):
        """⭐"""
        return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')



# -*- coding: utf-8 -*-
from logging import getLogger
from typing import Final

from aiogram import Bot, Router
from aiogram.exceptions import TelegramBadRequest, TelegramRetryAfter
from aiogram.types import ErrorEvent

from vpnflow.settings import settings

router: Final[Router] = Router(name=__name__)
logger = getLogger(__name__)
telegram_settings = settings.telegram


@router.errors()
async def error_handler(event: ErrorEvent, bot: Bot) -> None:
    """⭐"""
    logger.critical(event, exc_info=True)

    message = event.update.message

    if not message:
        return

    await message.answer(telegram_settings.messages["error"])

    user_id = message.from_user.id if message.from_user else ""
    text = telegram_settings.messages["error-notify"].format(
        user_id=user_id, error=event.exception
        )

    logger.error(text)

    if telegram_settings.log_chat:
        try:
            await bot.send_message(chat_id=telegram_settings.log_chat, text=text)
        except (TelegramBadRequest, TelegramRetryAfter) as exc:
            logger.error(exc)



# -*- coding: utf-8 -*-
from aiogram.types import BotCommand

from vpnflow.enums import BotCommands, BotCommandsAdmin
from vpnflow.settings import settings


def build_menu_from_enum(enum, commands_text=settings.telegram.messages["commands"]):
    """⭐"""
    for k, _ in enum.__members__.items():
        yield BotCommand(command=f"/{k}", description=commands_text[k])


user = list(build_menu_from_enum(BotCommands))
admin = user + list(build_menu_from_enum(BotCommandsAdmin))



# -*- coding: utf-8 -*-
import asyncio
from logging import getLogger

from aiogram.client.session.aiohttp import AiohttpSession
from aiogram.exceptions import TelegramRetryAfter

from vpnflow.bot import keyboards as kbs
from vpnflow.bot.base import create_bot
from vpnflow.db.base import engine_mrzbn, engine_r, run_session
from vpnflow.db.tools import (get_soon_expired_users,
                              get_telegram_expired_users, load_external_table)
from vpnflow.misc import days_verbose
from vpnflow.services._telegram import broadcast
from vpnflow.settings import settings

logger = getLogger(__name__)


async def _send_messsages(session, soon_expired_telegram_users):
    """⭐"""
    bot = create_bot(session=session)
    message_template = settings.telegram.messages["flows-notify-users"]
    async for days_exp, telegram_ids in soon_expired_telegram_users:
        message = message_template.format(
            days_exp=days_exp, days_exp_say=days_verbose(days_exp)
            )
        await broadcast(bot, message, *telegram_ids, reply_markup=kbs.inline.flow_notify_users)
    try:
        await bot.session.close()
        # await bot.close()
    except TelegramRetryAfter as exc:
        logger.error(exc)
    except Exception as exc:
        logger.error(f"{type(exc)}. {exc}")


async def _notify_users():
    """⭐"""
    users_table = await load_external_table(engine_mrzbn, "users")
    async with \
        run_session(engine_mrzbn) as session_marzban, \
        run_session(engine_r) as session, \
        AiohttpSession() as http_session \
        :
        soon_expired_marzban_users = get_soon_expired_users(
            session_marzban, users_table, *settings.flows.notify_users_days
            )
        soon_expired_telegram_users = get_telegram_expired_users(session, soon_expired_marzban_users)
        await _send_messsages(http_session, soon_expired_telegram_users)


def notify_users():
    """⭐"""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(_notify_users())



# -*- coding: utf-8 -*-
from logging import getLogger
from typing import Any, Awaitable, Callable, Dict

from aiogram import BaseMiddleware
from aiogram.types import CallbackQuery, Message, TelegramObject

from vpnflow.bot import keyboards as kbs
from vpnflow.bot.callbacks import UserCallback
from vpnflow.services import business
from vpnflow.services._telegram import edit_callback
from vpnflow.settings import settings

logger = getLogger(__name__)
messages = settings.telegram.messages


class CheckAcceptMiddleware(BaseMiddleware):
    """⭐"""

    def __init__(
        self,
        callbacks = ("start", "pay", "setup", "promo", "invite", "help", "setup-platform", "setup-help")
        ) -> None:
        self.callbacks = callbacks

    async def __call__(
            self,
            handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
            event: TelegramObject,
            data: Dict[str, Any],
    ) -> Any:
        if isinstance(event, Message) and "/start" not in str(event.text):
            user = await business.get_user(telegram_id=event.from_user.id)
            if user and user.marzban_username is None:
                await event.answer(messages["start-welcome"], reply_markup=kbs.inline.user_welcome)
                return
            else:
                data["user"] = user
                if user is None:
                    logger.warning(f"User is None, {event.from_user.id}")
                    return
        if isinstance(event, CallbackQuery):
            callback_data = data.get("callback_data")
            if (
                callback_data and
                isinstance(callback_data, UserCallback) and
                callback_data.action in self.callbacks
            ):
                user = await business.get_user(telegram_id=event.from_user.id)
                if user and user.marzban_username is None:
                    await edit_callback(event, messages["start-welcome"], reply_markup=kbs.inline.user_welcome)
                    return
                else:
                    data["user"] = user
                    if user is None:
                        logger.warning(f"User is None, {event.from_user.id}")
                        return
        result = await handler(event, data)
        return result


class FilterNoUserMiddleware(BaseMiddleware):
    """⭐"""

    async def __call__(
            self,
            handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
            event: TelegramObject,
            data: Dict[str, Any],
    ) -> Any:
        if isinstance(event, (CallbackQuery, Message)) and event.from_user is None:
            return
        result = await handler(event, data)
        return result


ALL = (FilterNoUserMiddleware(), )



# -*- coding: utf-8 -*-
from asyncio import create_task
from logging import getLogger
from typing import Final

from aiogram import Bot, F, Router
from aiogram.enums import ChatType
from aiogram.filters import Command
from aiogram.fsm.context import FSMContext
from aiogram.types import CallbackQuery, Message

from vpnflow.bot import callbacks as cb
from vpnflow.bot import keyboards as kbs
from vpnflow.bot import states as st
from vpnflow.bot.filters import AdminFilter
from vpnflow.enums import BotCommandsAdmin
from vpnflow.services import _telegram, business
from vpnflow.services._telegram import edit_callback
from vpnflow.settings import settings

logger = getLogger(__name__)
router: Final[Router] = Router(name=__name__)
router.message.filter(AdminFilter(), F.chat.type == ChatType.PRIVATE)
router.callback_query.filter(AdminFilter(), F.message.chat.type == ChatType.PRIVATE)

admins_id = settings.telegram.admins_id
messages = settings.telegram.messages


@router.message(Command(BotCommandsAdmin.panel.name))
async def cmd_panel(message: Message) -> None:
    """⭐"""
    await message.answer(
        messages[BotCommandsAdmin.panel.name], reply_markup=kbs.inline.admin
        )


@router.callback_query(
    cb.AdminCallback.filter(F.action == "panel"), F.from_user.id.in_(admins_id)
    )
async def cb_panel(callback_query: CallbackQuery) -> None:
    """⭐"""
    await edit_callback(
        callback_query, messages[BotCommandsAdmin.panel.name], reply_markup=kbs.inline.admin
        )


@router.callback_query(
    cb.AdminCallback.filter(F.action == "stats"), F.from_user.id.in_(admins_id)
    )
async def cb_stats(callback_query: CallbackQuery, bot: Bot) -> None:
    """⭐"""
    await callback_query.answer(messages["admin-stats-answr"])
    create_task(business.send_stats(bot, callback_query.from_user.id))
    await edit_callback(
        callback_query, messages["admin-stats-preview"],
        reply_markup=kbs.inline.admin_start
        )
    return


@router.callback_query(
    cb.AdminCallback.filter(F.action == "notify"), F.from_user.id.in_(admins_id)
    )
async def cb_notify(callback_query: CallbackQuery, state: FSMContext) -> None:
    """⭐"""
    await edit_callback(
        callback_query, messages["admin-notify"], reply_markup=kbs.inline.admin_notify
        )
    await state.set_state(st.NotifyForm.group_choice)


@router.callback_query(
    cb.NotifyCallback.filter(), F.from_user.id.in_(admins_id),
    st.NotifyForm.group_choice
    )
async def cb_choice_group(
    callback_query: CallbackQuery,
    callback_data: cb.NotifyCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    notify_group = callback_data.value
    await state.update_data(notify_group=notify_group)
    await edit_callback(
        callback_query, messages["admin-notify-text"]
        )
    await state.set_state(st.NotifyForm.message)


@router.message(st.NotifyForm.message)
async def notify_accept(message: Message, state: FSMContext) -> None:
    """⭐"""
    if message.photo is not None:
        await state.update_data(notify_image=message.photo[-1].file_id)
    await state.update_data(notify_message=message.html_text.strip())
    await message.answer(messages["admin-notify-accept"], reply_markup=kbs.inline.yes_no_menu)
    await state.set_state(st.NotifyForm.accept_choice)


@router.callback_query(
    cb.YesNoCallback.filter(), F.from_user.id.in_(admins_id),
    st.NotifyForm.accept_choice
    )
async def cb_accept_choice(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext, bot: Bot
    ) -> None:
    """⭐"""
    state_data = await state.get_data()
    await _telegram.clean_state(state)
    if callback_data.value:
        await callback_query.answer(messages["admin-notify-start"])
        notified_users = await business.notify_users(bot, **state_data)
        await edit_callback(
            callback_query,
            messages["admin-notify-end"].format(notified_users=notified_users),
            reply_markup=kbs.inline.admin_start
            )
    else:
        await edit_callback(
             callback_query, messages[BotCommandsAdmin.panel.name], reply_markup=kbs.inline.admin
             )


@router.callback_query(
    cb.AdminCallback.filter(F.action == "coupons"), F.from_user.id.in_(admins_id)
    )
async def cb_coupons(callback_query: CallbackQuery, state: FSMContext) -> None:
    """⭐"""
    await edit_callback(callback_query, messages["admin-coupons"])
    await state.set_state(st.CouponAddForm.name)


@router.message(st.CouponAddForm.name)
async def coupons_name(message: Message, state: FSMContext) -> None:
    """⭐"""
    await state.update_data(coupon_id=message.text.strip().upper())
    await message.answer(messages["admin-coupons-days"])
    await state.set_state(st.CouponAddForm.days)


@router.message(st.CouponAddForm.days)
async def coupons_accept(message: Message, state: FSMContext) -> None:
    """⭐"""
    await state.update_data(coupon_days=message.text.strip())
    await message.answer(messages["admin-coupons-usage-accept"], reply_markup=kbs.inline.yes_no_menu)
    await state.set_state(st.CouponAddForm.accept_usage_limit)


@router.callback_query(
    cb.YesNoCallback.filter(), F.from_user.id.in_(admins_id),
    st.CouponAddForm.accept_usage_limit
    )
async def coupons_accept_usage_limit(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    if callback_data.value:
        await edit_callback(callback_query, messages["admin-coupons-usage"])
        await state.set_state(st.CouponAddForm.usage_limit)
    else:
        await edit_callback(callback_query, messages["admin-coupons-accept"], reply_markup=kbs.inline.yes_no_menu)
        await state.set_state(st.CouponAddForm.accept_choice)


@router.message(st.CouponAddForm.usage_limit)
async def coupons_usage_limit(message: Message, state: FSMContext) -> None:
    """⭐"""
    await state.update_data(coupon_usage_limit=message.text.strip())
    await message.answer(messages["admin-coupons-accept"], reply_markup=kbs.inline.yes_no_menu)
    await state.set_state(st.CouponAddForm.accept_choice)


@router.callback_query(
    cb.YesNoCallback.filter(), F.from_user.id.in_(admins_id),
    st.CouponAddForm.accept_choice
    )
async def coupons_accept_choice(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    state_data = await state.get_data()
    await _telegram.clean_state(state)
    if callback_data.value:
        coupon_created = await business.create_coupon(**state_data)
        if coupon_created:
            await edit_callback(
                callback_query,
                messages["admin-coupons-success"],
                reply_markup=kbs.inline.admin_start
                )
        else:
            await edit_callback(
                callback_query,
                messages["admin-coupons-failed"],
                reply_markup=kbs.inline.admin_start
                )
    else:
        await edit_callback(
            callback_query, messages[BotCommandsAdmin.panel.name],
            reply_markup=kbs.inline.admin
            )

@router.callback_query(
    cb.AdminCallback.filter(F.action == "services"), F.from_user.id.in_(admins_id)
    )
async def cb_services(callback_query: CallbackQuery) -> None:
    """⭐"""
    await edit_callback(
        callback_query, messages["admin-services"],
        reply_markup=kbs.inline.admin_start
        )


@router.callback_query(
    cb.AdminCallback.filter(F.action == "ssh-add_marzban_node_service"), F.from_user.id.in_(admins_id)
    )
async def cb_ssh_add_node(callback_query: CallbackQuery, state: FSMContext) -> None:
    """⭐"""
    await edit_callback(callback_query, messages["admin-ssh-add_marzban_node_service"])
    await state.set_state(st.AddNode.data)


@router.message(st.AddNode.data)
async def ssh_add_node(message: Message, state: FSMContext) -> None:
    """⭐"""
    await state.update_data(node_data=message.text.strip())
    await message.answer(messages["admin-ssh-add_marzban_node_service-accept"], reply_markup=kbs.inline.yes_no_menu)
    await state.set_state(st.AddNode.accept_choice)


@router.callback_query(
    cb.YesNoCallback.filter(), F.from_user.id.in_(admins_id),
    st.AddNode.accept_choice
    )
async def cb_ssh_add_node_accept_choice(
    callback_query: CallbackQuery,
    callback_data: cb.YesNoCallback,
    state: FSMContext
    ) -> None:
    """⭐"""
    state_data = await state.get_data()
    await _telegram.clean_state(state)
    if callback_data.value:
        msg_code = await business.add_node(state_data)
        await edit_callback(
            callback_query,
            messages[msg_code] if msg_code in messages else messages["error"],
            reply_markup=kbs.inline.admin_start
            )
    else:
        await edit_callback(
            callback_query, messages[BotCommandsAdmin.panel.name],
            reply_markup=kbs.inline.admin
            )



# -*- coding: utf-8 -*-
from contextlib import asynccontextmanager
from logging import getLogger

from sqlalchemy import MetaData
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import (AsyncAttrs, AsyncEngine, AsyncSession,
                                    async_sessionmaker, create_async_engine)
from sqlalchemy.orm import DeclarativeBase

from vpnflow.settings import settings

logger = getLogger(__name__)
db_settings = settings.database


class Base(AsyncAttrs, DeclarativeBase):
    """⭐"""

    metadata: MetaData = MetaData()

    __abstract__ = True

    @classmethod
    def build_tablename(cls) -> str:
        """⭐"""
        return cls.__name__.lower() + 's'

    def __repr__(self) -> str:
        cols = [
            f"{col}={getattr(self, col)}"
            for idx, col in enumerate(self.__table__.columns.keys())
        ]
        return f"{self.__class__.__name__}({', '.join(cols)})"


def create_engine(
    url: str = db_settings.sqlalchemy_url.get_secret_value(),
    params = db_settings.sqlalchemy_engine_params
    ) -> AsyncEngine:
    """⭐"""
    return create_async_engine(url, **params)


def create_session_pool(
    engine: AsyncEngine,
    **kwargs: dict
    ) -> async_sessionmaker[AsyncSession]:
    """⭐"""
    expire_on_commit = kwargs.get("expire_on_commit", False)
    return async_sessionmaker(
        engine, class_=AsyncSession, expire_on_commit=expire_on_commit
        )


@asynccontextmanager
async def run_session(engine, **kwargs):
    """⭐"""
    session_runner = create_session_pool(engine)
    async with session_runner() as session:
        try:
            yield session
            if kwargs.get("commit", False):
                await session.commit()
        except (SQLAlchemyError, Exception) as exc:
            logger.error(exc)
            await session.rollback()
        finally:
            await session.close()


engine_r = create_engine()
engine_w = engine_r.execution_options(isolation_level="AUTOCOMMIT")

if settings.marzban.sqlalchemy_url:
    engine_mrzbn = create_engine(settings.marzban.sqlalchemy_url.get_secret_value())
else:
    engine_mrzbn = None



# -*- coding: utf-8 -*-
from functools import lru_cache

from aiogram.types import WebAppInfo
from aiogram.utils.keyboard import InlineKeyboardBuilder, InlineKeyboardMarkup

from vpnflow.bot import callbacks as cb
from vpnflow.settings import settings

buttons_text = settings.telegram.messages["buttons"]


def create_inline_keyboard(*buttons_data, rows_num=None, cols_num=None, attach=None, as_markup=True):
    """⭐"""
    kb = InlineKeyboardBuilder()
    for button_text, button_callback in buttons_data:
        kb.button(text=button_text, callback_data=button_callback)
    if attach:
        kb.attach(attach)
    if rows_num:
        if cols_num:
            kb.adjust(rows_num, cols_num)
        else:
            kb.adjust(rows_num)
    if as_markup:
        return kb.as_markup()
    return kb


@lru_cache(maxsize=32)
def build_referral(url: str) -> InlineKeyboardMarkup:
    """⭐"""
    kb = InlineKeyboardBuilder()
    kb.button(text=buttons_text["referral-share"], url=f"https://telegram.me/share/url?url={url}")
    kb.button(text=buttons_text["start"], callback_data=cb.UserCallback(action="start"))
    kb.adjust(1)
    return kb.as_markup()


@lru_cache(maxsize=32)
def build_setup(url: str) -> InlineKeyboardMarkup:
    """⭐"""
    kb = InlineKeyboardBuilder()
    kb.button(text=buttons_text["setup-open"], web_app=WebAppInfo(url=url))
    return kb.as_markup()


async def build_pay_plans(pay_plans_agenerator) -> InlineKeyboardMarkup:
    """⭐"""
    kb_data = []
    async for pay_plan in pay_plans_agenerator:
        kb_data.append(
            (
                pay_plan.button_text,
                cb.PayPlanCallback(action="choice-pay-plan", value=pay_plan.id))
                )
    kb_data.append((buttons_text["back"], cb.UserCallback(action="start")))
    return create_inline_keyboard(*kb_data, rows_num=1)


def build_pay_plan_prices(pay_plan) -> InlineKeyboardMarkup:
    """⭐"""
    kb_data = []
    for pay_plan_price in pay_plan.prices:
        kb_data.append(
            (
                pay_plan_price.currency.symbol,
                cb.PayPlanCallback(action="choice-pay-method", value=pay_plan_price.payment_method))
            )
    return create_inline_keyboard(*kb_data, rows_num=2)


@lru_cache(maxsize=32)
def build_setup_steps(download_url: str, add_url: str) -> InlineKeyboardMarkup:
    """⭐"""
    kb = InlineKeyboardBuilder()
    kb.button(text=buttons_text["setup-download"], url=download_url)
    kb.button(text=buttons_text["setup-add"], url=add_url)
    kb.button(text=buttons_text["start"], callback_data=cb.UserCallback(action="start"))
    kb.adjust(1)
    return kb.as_markup()


support = InlineKeyboardBuilder()
support.button(text=buttons_text["help-faq"], web_app=WebAppInfo(url=settings.business.url_faq))
support.button(text=buttons_text["help-ask"], url=settings.business.url_support)
support.button(text=buttons_text["setup-help"], callback_data=cb.UserCallback(action="setup-help"))
support.button(text=buttons_text["back"], callback_data=cb.UserCallback(action="start"))
support.adjust(1)
support = support.as_markup()

support_back = InlineKeyboardBuilder()
support_back.button(text=buttons_text["back"], callback_data=cb.UserCallback(action="help"))
support_back.adjust(1)
support_back = support_back.as_markup()

yes_no_menu = create_inline_keyboard(
    *(
        (buttons_text["choice-yes"], cb.YesNoCallback(value=True)),
        (buttons_text["choice-no"], cb.YesNoCallback(value=False))
    ), rows_num=2
)

user_promo = create_inline_keyboard(
    *(
        (buttons_text["promo-again"], cb.UserCallback(action="promo")),
        (buttons_text["start"], cb.UserCallback(action="start"))
    ), rows_num=1
)

user_welcome = InlineKeyboardBuilder()
user_welcome.button(text=buttons_text["start-accept"], callback_data=cb.UserCallback(action="sign_up"))
user_welcome.adjust(1)
user_welcome = user_welcome.as_markup()

user_start = InlineKeyboardBuilder()
user_start.button(text=buttons_text["start"], callback_data=cb.UserCallback(action="start"))
user_start = user_start.as_markup()

user_start_new = InlineKeyboardBuilder()
user_start_new.button(text=buttons_text["start-setup"], callback_data=cb.UserCallback(action="setup"))
user_start_new = user_start_new.as_markup()

user_setup_platform = create_inline_keyboard(
    *(
        (
            platform, cb.UserCallback(action="setup-platform", value=platform)
            ) for platform in settings.business.supported_platforms
    ), rows_num=4
)

user = create_inline_keyboard(
    *(
        (
            buttons_text[k], cb.UserCallback(action=k)
            ) for k in ("pay", "setup", "promo", "invite", "help")
    ), rows_num=1
)


admin = create_inline_keyboard(
    *(
        (
            buttons_text[f"admin-{k}"], cb.AdminCallback(action=k)
            ) for k in ("stats", "notify", "coupons", "services", "ssh-add_marzban_node_service")
    ), rows_num=1
)

admin_notify = create_inline_keyboard(
    *(
        (buttons_text["admin-notify-all"], cb.NotifyCallback(value="all")),
        (buttons_text["admin-notify-active"], cb.NotifyCallback(value="active")),
        (buttons_text["admin-notify-expired"], cb.NotifyCallback(value="expired")),
        (buttons_text["admin-notify-pay-expired"], cb.NotifyCallback(value="pay-expired")),
        (buttons_text["admin-notify-pay-no-expired"], cb.NotifyCallback(value="pay-no-expired")),
        (buttons_text["back"], cb.AdminCallback(action="panel"))
    ), rows_num=1
)

admin_start = InlineKeyboardBuilder()
admin_start.button(text=buttons_text["panel"], callback_data=cb.AdminCallback(action="panel"))
admin_start = admin_start.as_markup()


if buttons_text.get("flows-notify-users") is None:
    flow_notify_users = None
else:
    flow_notify_users = InlineKeyboardBuilder()
    flow_notify_users.button(text=buttons_text["flows-notify-users"], callback_data=cb.UserCallback(action="pay"))
    flow_notify_users = flow_notify_users.as_markup()



import sys
import time
from logging import getLogger

import schedule

from vpnflow.flows.base import notify_users
from vpnflow.settings import settings

logger = getLogger(__name__)


def scheduled_notify_users():
    """⭐"""
    try:
        notify_users()
    except Exception as exc:
        logger.error(exc)


def _get_time_from_cron(s: str):
    """⭐"""
    m, h, *_ = s.split()
    return f"{h}:{m}"


def run_scheduled_tasks(wait_time: int = 60):
    """⭐"""
    notify_users_time = _get_time_from_cron(settings.flows.notify_users_cron)
    schedule.every().day.at(notify_users_time).do(scheduled_notify_users).tag("daily", "notify")
    try:
        while True:
            logger.info(schedule.get_jobs())
            schedule.run_pending()
            time.sleep(wait_time)
    except (KeyboardInterrupt, SystemExit):
        schedule.clear()
    finally:
        sys.exit()



# -*- coding: utf-8 -*-
import time
from contextlib import contextmanager
from logging import getLogger

from paramiko import AutoAddPolicy, SSHClient

logger = getLogger(__name__)


@contextmanager
def ssh_session(host, username, password):
    """⭐"""
    logger.info(f"Start ssh session {host}@{username}")
    with SSHClient() as ssh:
        t0 = time.perf_counter()
        ssh.set_missing_host_key_policy(AutoAddPolicy())
        try:
            ssh.connect(host, username=username, password=password)
            yield ssh
        except Exception as exc:
            logger.error(f"Exception in ssh session: {exc}")
        finally:
            logger.info(f"Ssh session duration: {time.perf_counter() - t0 :.2f} s")


def upload_files(ssh_session, *files_paths):
    """⭐"""
    with ssh_session.open_sftp() as sftp:
        for local_path, remote_path in files_paths:
            try:
                sftp.put(local_path, remote_path)
            except FileNotFoundError as exc:
                logger.error(exc)
                return False
            else:
                logger.debug(f"File uploaded {local_path} -> {remote_path}")
    return True


def exec_commands(ssh_session, *commands, timeout=600):
    """⭐"""
    for command in commands:
        logger.debug(f"Execute command: {command}")
        stdin, stdout, stderr = ssh_session.exec_command(command, timeout=timeout)
        logger.debug("Stdout:\n{stdout}".format(stdout='\n'.join(l.strip() for l in stdout)))


def add_marzban_node_service(host, username, password, files_paths, commands):
    """⭐"""
    with ssh_session(host, username, password) as session:
        is_files_uploaded = upload_files(session, *files_paths)
        if not is_files_uploaded:
            logger.error(f"Files not uploaded: {files_paths}")
            return
        exec_commands(session, *commands)



# -*- coding: utf-8 -*-
from logging import getLogger
from ssl import create_default_context

import certifi

from vpnflow.settings import settings

logger = getLogger(__name__)
ssl_context = create_default_context(cafile=certifi.where())


class Cloudflare:
    """⭐"""

    settings = settings.cloudflare

    is_ready = all(
        (
            settings.email.get_secret_value(),
            settings.api_key.get_secret_value(),
            settings.zone_id.get_secret_value()
            )
        )

    @classmethod
    async def request(
        cls,
        session,
        method,
        path,
        json,
        headers={
            "Content-Type": "application/json",
            "X-Auth-Email": settings.email.get_secret_value(),
            "Authorization": f"Bearer {settings.api_key.get_secret_value()}"
            }
        ):
        """⭐"""
        session_request = session.request(
            method, cls.settings.host_api + path,
            json=json, headers=headers, ssl=ssl_context
            )
        async with session_request as resp:
            if resp.status == 200:
                resp_data = await resp.json()
                return resp_data
            else:
                try:
                    resp_data = await resp.json()
                except Exception:
                    resp_data = await resp.text()
                logger.error(f"Status: {resp.status}. Message: {resp_data}")
                resp.raise_for_status()



# -*- coding: utf-8 -*-
import json
import logging
from asyncio import gather
from asyncio import sleep as asleep

from aiogram import Bot
from aiogram.types.labeled_price import LabeledPrice

from vpnflow.misc import slice_generator
from vpnflow.settings import settings
from vpnflow.web.schemas import (YookassaProviderData,
                                 YookassaProviderDataReceipt,
                                 YookassaProviderDataReceiptItem)

logger = logging.getLogger(__name__)
settings_telegram = settings.telegram


def create_ref_link(start: str, bot_name: str = settings_telegram.bot_name):
    """⭐"""
    return f"https://t.me/{bot_name}?start={start}"


async def send_invoice(callback_query, payment):
    """⭐"""
    price, currency, title, payload = (
        payment.price, payment.currency, payment.title, str(payment.id)
    )
    description = settings_telegram.messages["pay-invoice"]
    is_telegram_pay = currency == "XTR"
    if is_telegram_pay:
        label = currency
    else:
        label = "Оплата"
        price *= 100
    prices = [LabeledPrice(label=label, amount=price)]
    if is_telegram_pay:
        await callback_query.message.answer_invoice(
            title=title, description=description, payload=payload,
            currency=currency, prices=prices # reply_markup=reply_markup
        )
    else:
        provider_token = settings_telegram.provider_token_yookassa.get_secret_value()
        need_email = settings_telegram.provider_need_email
        send_email_to_provider = settings_telegram.provider_send_email_to_provider
        provider_data = YookassaProviderData(
            receipt=YookassaProviderDataReceipt(
                items=[
                    YookassaProviderDataReceiptItem(
                        description=settings.business.default_desc,
                        amount={"value": payment.price, "currency": currency}
                        )
                    ],
                )
            )
        provider_data = json.dumps(provider_data.dict())
        await callback_query.bot.send_invoice(
            chat_id=callback_query.from_user.id, title=title, description=description,
            payload=payload, provider_token=provider_token, currency=currency, prices=prices,
            provider_data=provider_data, need_email=need_email,
            send_email_to_provider=send_email_to_provider # reply_markup=reply_markup
        )


async def clean_state(state) -> None:
    """⭐"""
    current_state = await state.get_state()
    if current_state is not None:
        await state.clear()


async def broadcast(bot: Bot, message: str, *recipients, photo=None, reply_markup=None):
    """⭐"""
    for sliced_recipients in slice_generator(recipients, 30):
        if photo:
            results = await gather(
                *(
                    bot.send_photo(
                        chat_id=recipient, caption=message, photo=photo
                        ) for recipient in sliced_recipients
                    ),
                return_exceptions=True
            )
        else:
            results = await gather(
                *(
                    bot.send_message(
                        recipient, message, reply_markup=reply_markup
                        ) for recipient in sliced_recipients
                    ),
                return_exceptions=True
            )
        for recipient, result in zip(sliced_recipients, results):
            logger.info(f"Broadcast recipient: {recipient}")
            if isinstance(result, Exception):
                logger.warning(str(result))
        await asleep(1)


async def edit_callback(
    callback_query, message, reply_markup=None, disable_web_page_preview=True
    ):
    """⭐"""
    try:
        await callback_query.message.edit_text(
            message, reply_markup=reply_markup, disable_web_page_preview=disable_web_page_preview
            )
    except Exception as exc:
        logger.warning(exc)
        callback_query.answer()
